# Generic Makefile to run tests of one or more programs
# Last edited on 2011-05-09 13:13:06 by stolfi

SHELL := /bin/bash

######################################################################
# To be included by the top-level Makefile in the program's directory.
# The caller must define
#
#   ${PROG} or ${PROGS}  test programs (whose sources are in current directory)
#   ${PROGDIR}           the directory containing the executable program and its sources
#   ${MY_LIB_DIR}        directory to search for ".a" files.
#   ${MY_INC_DIR}        directory to search for ".h" files.
#   
# The client must define the "all", "run", and "clean" targets.

ifeq "/${PROGS}" "/"
PROGS := ${PROG}
endif

ifneq "/${PROGS}" "/"
ifneq "/${PROGDIR}" "/"

#---------------------------------------------------------------------
# Phony "make" targets:
# Client hould define "all" and "clean"

.PHONY:: \
  default-action \
  all actions \
  build build-progs \
  clean

#---------------------------------------------------------------------
# Default target: call the client's "all" target.
default-action: all

#---------------------------------------------------------------------
# "make actions" performs the actions listed in ${ACTIONS}, if any
actions:
ifeq "/${ACTIONS}" "/"
	@echo '$${ACTIONS} is empty'
else
	${MAKE} ${ACTIONS}
endif

#---------------------------------------------------------------------
# "make build" and "make build-progs" make sure that the
# executables in ${PROGDIR} are up to date:

build: build-progs
  
PROG_XFILES := ${addprefix ${PROGDIR}/, ${PROGS}}

build-progs: ${PROG_XFILES}

${PROG_XFILES}: \
          ${wildcard ${PROGDIR}/*.h ${PROGDIR}/*.c} \
          ${wildcard ${MY_INC_DIR}/*.h} \
          ${wildcard ${MY_LIB_DIR}/*.a ${MY_LIB_DIR}/*.so}
	@echo "rebuilding ${PROG_XFILES} ..."
	cd ${PROGDIR}/. && ${MAKE} ${PROGS}

#---------------------------------------------------------------------
# "make clean" deletes all derived files.
# It is a double-colon rule so that clients may add more actions.

clean::
        
endif
endif
# End of ${PROG} ${PROGDIR} section.
######################################################################

